"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t;
    return { next: verb(0), "throw": verb(1), "return": verb(2) };
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var artillery_core_1 = require("artillery-core");
var aws = require("aws-sdk");
var fs = require("fs");
var uuid = require("uuid");
var sl = require("stats-lite");
var _ = require("lodash");
var csv = require("csv-parse");
aws.config.apiVersions = {
    lambda: '2015-03-31'
};
aws.config.update({
    region: process.env.AWS_REGION || 'us-west-2'
});
var sqs = new aws.SQS();
var lambda = new aws.Lambda({ httpOptions: { timeout: 340000 } });
var finishedLambdaCount = 0;
var finalSqsCount = 0;
var runId = uuid.v4();
function handler(args) {
    return __awaiter(this, void 0, void 0, function () {
        var config, lambdaPromises, receivedStats;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    config = readJsonFromFile(args.config);
                    return [4 /*yield*/, invokeLambdas(config, args)];
                case 1:
                    lambdaPromises = _a.sent();
                    return [4 /*yield*/, watchSqs(config, args)];
                case 2:
                    receivedStats = _a.sent();
                    return [4 /*yield*/, Promise.all(lambdaPromises)];
                case 3:
                    _a.sent();
                    console.log("\nAll lambdas finished processing.");
                    displayLatencyAggregates(receivedStats.latencies);
                    displayResponseCodeAggregates(receivedStats.responseCodes);
                    return [2 /*return*/];
            }
        });
    });
}
exports.handler = handler;
function invokeLambdas(conf, args) {
    return __awaiter(this, void 0, void 0, function () {
        var script, payloadCsv, payload, lambdaParams, lambdaPromises, i;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    script = readScriptFromFile(args.script);
                    return [4 /*yield*/, readPayloadFromFile(args.payloadCsv)];
                case 1:
                    payloadCsv = _a.sent();
                    payload = {
                        script: script,
                        payload: payloadCsv,
                        uid: runId
                    };
                    lambdaParams = constructLambdaParameters(conf, payload);
                    console.log("Invoking %s Lambdas", args.lambdaCount);
                    lambdaPromises = [];
                    for (i = 0; i < args.lambdaCount; i++) {
                        process.stdout.write('/');
                        lambdaPromises.push(lambda.invoke(lambdaParams).promise().then(function (response) {
                            finishedLambdaCount++;
                            args.intermediates && console.log("\nLambda finished execution", response.StatusCode);
                            process.stdout.write('-');
                            return response;
                        }));
                    }
                    return [2 /*return*/, lambdaPromises];
            }
        });
    });
}
function watchSqs(conf, args) {
    return __awaiter(this, void 0, void 0, function () {
        var continueWatching, receivedStats, params, intermediateResults;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    continueWatching = true;
                    receivedStats = {
                        messageIds: [],
                        latencies: [],
                        responseCodes: {}
                    };
                    _a.label = 1;
                case 1:
                    if (!continueWatching)
                        return [3 /*break*/, 3];
                    if (finalSqsCount >= args.lambdaCount) {
                        continueWatching = false;
                    }
                    process.stdout.write('.');
                    params = {
                        QueueUrl: conf.queueURL,
                        MaxNumberOfMessages: 10,
                        WaitTimeSeconds: 5
                    };
                    return [4 /*yield*/, sqs.receiveMessage(params).promise()];
                case 2:
                    intermediateResults = _a.sent();
                    if (intermediateResults && intermediateResults.Messages && intermediateResults.Messages.length > 0) {
                        intermediateResults.Messages.forEach(function (message) {
                            handleMessage(conf, args, message, receivedStats);
                        });
                    }
                    return [3 /*break*/, 1];
                case 3: return [2 /*return*/, receivedStats];
            }
        });
    });
}
function handleMessage(conf, args, message, receivedStats) {
    var body = JSON.parse(message.Body);
    var messagePayload = JSON.parse(body.Message);
    if (messagePayload.uid == runId) {
        var messageMarkedFinal = false;
        // SQS is at least once delivery, so we should handle duplicate messages.
        if (receivedStats.messageIds.indexOf(message.MessageId) == -1) {
            receivedStats.messageIds.push(message.MessageId);
            if (args.intermediates) {
                displayIntermediateResults(messagePayload);
            }
            else {
                process.stdout.write('*');
            }
            processIntermediateResults(messagePayload, receivedStats);
            messageMarkedFinal = messagePayload.type === 'final';
        }
        deleteSqsMessage(conf.queueURL, message.ReceiptHandle, messageMarkedFinal);
    }
    else {
        process.stdout.write('.');
    }
}
function deleteSqsMessage(sqsUrl, receiptHandle, final) {
    var params = {
        QueueUrl: sqsUrl,
        ReceiptHandle: receiptHandle
    };
    sqs.deleteMessage(params).promise()
        .then(function (response) {
        if (final) {
            finalSqsCount++;
        }
    })["catch"](function (error) { return console.log("Could not delete SQS message", receiptHandle, error, error.stack); });
}
function displayIntermediateResults(intermediateResults) {
    console.log("\nLambda Reported Intermediate Results:");
    displayLatencyAggregates(intermediateResults.stats._latencies);
    displayResponseCodeAggregates(intermediateResults.stats._codes);
}
function processIntermediateResults(intermediateResults, receivedStats) {
    if (intermediateResults.stats._codes) {
        for (var code in intermediateResults.stats._codes) {
            if (!receivedStats.responseCodes[code])
                receivedStats.responseCodes[code] = 0;
            receivedStats.responseCodes[code] += intermediateResults.stats._codes[code];
        }
    }
    if (intermediateResults.stats._latencies) {
        (_a = receivedStats.latencies).push.apply(_a, intermediateResults.stats._latencies);
    }
    var _a;
}
exports.processIntermediateResults = processIntermediateResults;
function constructLambdaParameters(conf, payload) {
    var params = {
        FunctionName: conf.lambdaName,
        InvocationType: 'RequestResponse',
        LogType: 'Tail',
        Payload: JSON.stringify(payload)
    };
    return params;
}
function readPayloadFromFile(payloadCSVFilename) {
    return __awaiter(this, void 0, void 0, function () {
        var filename, csvData;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!payloadCSVFilename) {
                        return [2 /*return*/, null];
                    }
                    filename = getFullFilePath(payloadCSVFilename);
                    return [4 /*yield*/, new Promise(function (resolve, reject) {
                            var data = [];
                            fs.createReadStream(filename).pipe(csv())
                                .on('data', function (row) {
                                data.push(row);
                            })
                                .on('error', function (e) {
                                reject(e);
                            })
                                .on('end', function () {
                                resolve(data);
                            });
                        })];
                case 1:
                    csvData = _a.sent();
                    return [2 /*return*/, csvData];
            }
        });
    });
}
exports.readPayloadFromFile = readPayloadFromFile;
function readScriptFromFile(scriptFilename) {
    var script = readJsonFromFile(scriptFilename);
    var validation = artillery_core_1.validate(script);
    if (validation.valid) {
        console.log("Script Passes Artillery Validation");
    }
    else {
        console.log("Artillery script did not validate", validation.errors);
        throw new Error("Artillery script did not validate");
    }
    return script;
}
exports.readScriptFromFile = readScriptFromFile;
function readJsonFromFile(scriptFilename) {
    var filename = getFullFilePath(scriptFilename);
    return JSON.parse(fs.readFileSync(filename, 'utf8'));
}
function displayResponseCodeAggregates(responseCodes) {
    for (var code in responseCodes) {
        console.log('  %s: %s', code, responseCodes[code]);
    }
}
function displayLatencyAggregates(latencies) {
    var min = Math.round(_.min(latencies) / 1e6);
    var max = Math.round(_.max(latencies) / 1e6);
    var mean = Math.round(sl.mean(latencies) / 1e6);
    var median = Math.round(sl.median(latencies) / 1e6);
    var p95 = Math.round(sl.percentile(latencies, 0.95) / 1e6);
    var p99 = Math.round(sl.percentile(latencies, 0.99) / 1e6);
    console.log('  min:  %s', min);
    console.log('  max:  %s', max);
    console.log('  mean: %s', mean);
    console.log('  p50:  %s', median);
    console.log('  p95:  %s', p95);
    console.log('  p99:  %s', p99);
}
function getFullFilePath(fileName) {
    return fileName.startsWith("\/") ? fileName : process.cwd() + '/' + fileName;
}
